// mpicc Ex4.c -Wall -o Ex4.x

#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

#include "par_child.h"

#define MIN(x, y) (x) < (y) ? (x) : (y)

// For simplicity, we assume there is an actual recvbuf for all processes
//  and not only for the root.
int myMPI_Reduce( const void *sendbuf, void *recvbuf, int count,
                  int root, MPI_Comm comm )
{
    int myrank, nprocs;
    int tag = 0;
    int i;

    int parent, left_ch, right_ch;
    double *tmp_buffer;
    

    MPI_Comm_rank( comm, &myrank );
    MPI_Comm_size( comm, &nprocs );

    get_parent_and_children( myrank, root, nprocs,
                             &parent, &left_ch, &right_ch );

    // Temporary buffer to accumulate results
    tmp_buffer = (double *) malloc ( count * sizeof(double) );

    // We copy send into the temporary buffer to simplify the calculations below.
    memcpy( tmp_buffer, sendbuf, count*sizeof(double) );

    if ( left_ch != -1 )
    {
        // If I have a left child, receive its message
        MPI_Recv( recvbuf, count, MPI_DOUBLE, left_ch, tag, comm, MPI_STATUS_IGNORE );
        for ( i = 0; i < count; i++ )
            tmp_buffer[i] = MIN(((double *)recvbuf)[i], tmp_buffer[i]);
    }
    if ( right_ch != -1 )
    {
        // If I have a right child, receive its message
        MPI_Recv( recvbuf, count, MPI_DOUBLE, right_ch, tag, comm, MPI_STATUS_IGNORE );
        for ( i = 0; i < count; i++ )
            tmp_buffer[i] = MIN(((double *)recvbuf)[i], tmp_buffer[i]);
    }

    // Copy the temporary into recv and pass along
    memcpy( recvbuf, tmp_buffer, count*sizeof(double) );

    if ( parent != -1 )
    {
        // If I have a parent (i.e., I am not the root), pass the message along
        MPI_Send( recvbuf, count, MPI_DOUBLE, parent, tag, comm );
    }

    free( tmp_buffer );

    return 0;
}

int main( int argc, char *argv[] )
{   
    int i;
    int ROOT = 26;
    int myrank, nprocs;
    double sendbuf[1], recvbuf[1];

    MPI_Init(&argc,&argv);

    MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
    MPI_Comm_size( MPI_COMM_WORLD, &nprocs );
    
    srand( myrank*nprocs );
    sendbuf[ 0 ] = ((double)rand()) / RAND_MAX;
    recvbuf[ 0 ] = 0;
        
    for ( i = 0; i < 1; i++ )
        printf( "[%d] %d: %f\n", myrank, i, sendbuf[i] );
    fflush(stdout);
    
    myMPI_Reduce( (const void *)sendbuf, (void *)recvbuf, 1, ROOT, MPI_COMM_WORLD );

    MPI_Barrier( MPI_COMM_WORLD );

    if ( myrank == ROOT )
    {
        printf( "MIN:\n" );
        for ( i = 0; i < 1; i++ )
            printf( "[%d] %d: %f\n", myrank, i, recvbuf[i] );
    }

    MPI_Finalize();
    return 0;
}


